package project.model;
import becker.util.IView;
import becker.io.*;
import java.awt.Polygon;

/** Landscape Design Planner application logic.
 @author Sandy Graham
 @version 1.0 25-Sep-2003 */
public class LandscapePlanner extends Object
{

   /** last thing we did */
   private String lastSelection = "";

   /** current action chosen */
   private String action = this.getItemList()[0];

   /** A partially filled array to keep track of the views used by this
    model. */
   private static int MIN_NUM_VIEWS = 4;
   private IView[] views = new IView[MIN_NUM_VIEWS];
   private int numViews = 0;

   private Rect plantingSquare;  // the square that things will be planted into
   private Rect selectSquare; // a temporary square used in updatedata

   private SquareContents activePlant; // the square contents that will be put into the selected area
   private SquareList squareList = new SquareList("plantlist.txt"); // the list of possible square contents

   private Grid[] theLayers = new Grid[Constants.NUM_LAYERS]; // the base and plant layers
   private Grid[][] undoLayers = new Grid[Constants.NUM_LAYERS][Constants.NUM_UNDO_LEVELS]; // the undo layers.  One dimension is the undo's, the other is the 2 layers
   private boolean baseIsFull = false; // whether or not the base layer is full yet
   private int curLayer; // the current layer the user is looking at
   private int curAction;  //The current action taking place ie.remove, placing stuff

   private Polygon myGon = new Polygon(); // the polygon the user is selecting stuff with

   private boolean showingShadow = false;
   private int displayHour = Constants.HOURS_OF_SUN; // if it is 14, we are display all shadows throughout the day.  Otherwise, it is that hour's shadow
   private CanopyLayer myCanLayer;

   private String curQuery; // the current query
   /** Construct the layout designer.
    @param gridWidth the width of the design palette area
    @param gridHeight the height of the design palette area
    */

    /** Returns if we are showing the shadow layer right now
     @return whether or not we are showing shadow
     */
    public boolean getShowingShadow()
    {  return this.showingShadow;
    }
    /** This function returns either the total sunlight hours or it returns a value indicating if there is sunlight at a given hour
    * @return the hours of sunlight.  Essentially the number is used in color calculations in the gridView
    */
    public int getSun(int x,int y)
    {  if(this.displayHour == 14)
       {   return this.myCanLayer.getHoursOfSun(x,y); // this simply returns the number of hours of sun
       }
       else
       {   return (14 - ((myCanLayer.getShadowAtHour(x,y,displayHour))) * 7); // getShadowAtHour returns 0 or 1 depending on the hour.  So, we multiply it by 7 (to get a nice grey) and subtract that from 14 (to get white in the absence of shadow)

       }
    }
    /** Move all the undo memories back one, and copy the current grids into the highest one
    */
    private void saveGrids()
    {   for(int x= 0 ;x < Constants.NUM_UNDO_LEVELS-1;x++)
        {   for(int y =0;y < Constants.NUM_LAYERS;y++)
            {   undoLayers[y][x] = undoLayers[y][x+1];
            }
        }
        for(int y = 0;y < Constants.NUM_LAYERS;y++)
        {   undoLayers[y][Constants.NUM_UNDO_LEVELS-1] = new Grid(this.theLayers[y]);
        }
    }
    /** Set the current grids equal to the highest undo grid, and move everything up one */
    public void undo()
    {   for(int y = 0; y< Constants.NUM_LAYERS;y++)
        {   this.theLayers[y] = undoLayers[y][Constants.NUM_UNDO_LEVELS-1];
        }

        for(int x = Constants.NUM_UNDO_LEVELS-1;x > 0;x--)
        {  for(int y = 0; y < Constants.NUM_LAYERS;y++)
           {    this.undoLayers[y][x] = this.undoLayers[y][x-1];
           }
        }
        Debug.O("undoing...");
        this.updateAllViews();
    }
    /** returns if conditions are correct for an undo operation.  Called by commandview to enable/disable the undo button */
    public boolean undoAble()
    {   for(int y = 0;y < Constants.NUM_LAYERS-1;y++)
        {   if(this.undoLayers[y][Constants.NUM_UNDO_LEVELS-1] == null)
            {   return false;
            }
        }
        return true;
    }
    /** constructs the landscape planner to contain all blank squares, and to have a gridWidth by gridHeight grid
    * @param gridWidth the width of the grid
    * @param gridHeight the height of the grid
    */
   public LandscapePlanner(int gridWidth, int gridHeight)
   {  super();
      theLayers[Constants.BASE_LAYER] = new Grid(20,20);
      theLayers[Constants.PLANT_LAYER] = new Grid(20,20,true);
      this.curAction = Constants.ACTION_PLANT_NORMAL;
      activePlant = new InorganicContents();
      this.myCanLayer = new CanopyLayer(theLayers);
   } // LandscapePlanner

   /** Constructs the landscape planner from a textfile of format described on the internet
   * @param in this is the textinput object initialized to the file you want to read
   */
   public LandscapePlanner(TextInput in)
   {  super();
      int gridWidth = in.readInt();
      int gridHeight = in.readInt();
      in.readLine();
      for(int x = 0;x < 2; x++)
      {   theLayers[x] = new Grid(gridWidth, gridHeight, in, squareList);
      }
      theLayers[Constants.PLANT_LAYER].loadSoilMulch(in,gridWidth,gridHeight);

      this.baseIsFull = theLayers[Constants.BASE_LAYER].isFull();
      activePlant = new InorganicContents();
      calcWater();
      this.myCanLayer = new CanopyLayer(theLayers);
      this.updateAllViews();
   } // LandscapePlanner from file
   /** This returns the added up install cost from the plant layer
   * @return the added up install cost from the plant layer
   */
   public int getInstallCost()
   {  return this.theLayers[Constants.PLANT_LAYER].getInstallCost();
   }
   /** This returns the added up maintanance cost for the plant layer for one year
   * @return the maintanance cost for the plant layer
   */
   public int getMaintainCost()
   {  return this.theLayers[Constants.BASE_LAYER].getMaintainCost();
   }
   // Accessor methods for the layout designer
   /** Accessor method
   * @return the selected square
   */
   public Rect getSelectSquare()
   {
       return this.selectSquare;
   }
   /** This updates the active selection (which is what is highlighted).  It is called from the gridview, whenever the mouse moves to a new grid square
   * @param x the x co-ordinate of the updated data
   * @param y the y co-ordinate of the updated data
   */
   public void setCurActiveSection(int x,int y)
   {
      if(this.plantingSquare != null)
      {    this.selectSquare = new Rect(this.plantingSquare.getLeft(),this.plantingSquare.getTop());
           this.selectSquare.setLeft(plantingSquare.getLeft());
           this.selectSquare.setTop(plantingSquare.getTop());
           this.selectSquare.setBottom(y);
           this.selectSquare.setRight(x);
           Debug.O("setting up selectSquare" + selectSquare.report());
           Debug.O("planting square: " + plantingSquare.report());
      }
      this.updateAllViews();
   }
   /** Gets the currently active plant (the plant/object next to be planted)
   * @return the currently active plant
   */
   public SquareContents getActivePlant()
   {
      return this.activePlant;
   }
   /** Return the items we can deal with depending on the current mode-- put in the drop-down menu. */
   public String[] getItemList()
   {  if(this.curLayer == Constants.BASE_LAYER)
      {  System.out.println("returning base layer actions");
         return new String[] {"Add Building", "Add Water", "Add Concrete", "Add Soil"};
      }
      else
      {   System.out.println("returning plant layer actions");
          return squareList.compileArray();
      }
   }
   /** Return the actions we can do (add, remove, etc.) in the current mode
   * @return an array of strings containing all we can do
   */
   public String[] getActionList()
   {   if(this.curLayer == Constants.PLANT_LAYER)
       {   return new String [] {"Add Normal","Add Polygon", "Remove","Query","Add Mulch","Remove Mulch","Add Topsoil","Remove Topsoil"};
       }
       else
       {   return new String[] {"Add Normal","Add Polygon", "Remove","Query"};
       }
   }

   /** Return the width of the design area. */
   public int getWidth()
   {  return Constants.DEFAULT_WIDTH;
   } // getWidth

   /** Return the height of the design area. */
   public int getHeight()
   {  return Constants.DEFAULT_HEIGHT;
   } // getHeight

	/** Get the data from a particular square in the grid.
	 * This method will likely need to change to provide more complete data
	 * and to NOT return a constant every time it's called! */
   public Square getData(int x, int y)
   {  return theLayers[this.curLayer].getSquare(x,y);
   }

   /** Update the action that will be performed by the user next.
    @param chosenAction indicates what the user will do next
    */
   public void setActionValue(String chosenAction)
   {  this.action = chosenAction;
      int squareNum = this.parseItem(this.action);
      this.activePlant = squareList.getSquare(squareNum);
      this.updateAllViews();
   }

   /** Parse the action string and return a square contents type
   */
   public int parseItem(String action)
   {  if (action.equalsIgnoreCase("Add Building"))
      { return 0;
      }
      else if (action.equalsIgnoreCase("Add Water"))
      { return 2;
      }
      else if (action.equalsIgnoreCase("Add Concrete"))
      { return 1;
      }
      else if (action.equalsIgnoreCase("Add Soil"))
      { return 3;
      }
      else if(squareList.search(action.substring(4)) >= 0)
      {  
         return squareList.search(action.substring(4));
      }
      else
      { return Constants.EMPTY_SQUARE;
      }

   }
   /** This function will throw stuff down when updatedata determines it is time to do so
   * @param whatToPut the squarecontents to place on the grid
   * @param where - where to put said squarecontents
   */
   public void placeStuff(SquareContents whatToPut,Rect where)
   {
      Debug.O("Now Placing a ");
      Debug.O(whatToPut.getName());
      Debug.O("-----------------");

      this.theLayers[this.curLayer].setRect(where,whatToPut, false);
      this.baseIsFull = this.theLayers[Constants.BASE_LAYER].isFull();

      if(whatToPut.equals(squareList.getSquare(Constants.NUM_SOIL)) && (this.curLayer == Constants.BASE_LAYER)  )
      {   this.theLayers[Constants.PLANT_LAYER].removeRect(where,squareList,this.theLayers[Constants.BASE_LAYER]);
      }

      this.calcWater();
      this.updateAllViews();
   }
   /** This returns an array of strings containing the possible layer switches.  It ends up in a combo box */
   public String[] getLayerList()
   {  return new String[] {"Base Layer","Plant Layer","Canopy Layer"};
   }
   /** switches the layer between base layer and plant layer (and hopefully, canopy layer)
   * @param layer - which layer we are switching to
   */
   public void switchLayer(String layer)
   {  if(layer.equalsIgnoreCase("Base Layer"))
      {   this.curLayer = Constants.BASE_LAYER;
          this.showingShadow = false;
      }
      else if(layer.equalsIgnoreCase("Plant Layer"))
      {   this.curLayer = Constants.PLANT_LAYER;
          this.showingShadow = false;
      }
      else
      {   this.myCanLayer = new CanopyLayer(theLayers);
          this.showingShadow = true;
      }
      this.updateAllViews();
   }
   /** returns if the base layer is full yet */
   public boolean getFull()
   {  return this.baseIsFull;
   }


   /** @param x the x-coordinate
    @param y the y-coordinate
    @return true if the coordinates are within the array boundaries
    */
   private boolean isValidCoordinate(int x, int y)
   {  return (0 <= x && x < this.getWidth()) && (0 <= y && y < this.getHeight());
   } // isValidCoordinate


   /** Checks if a given x,y co-ord is already in our polygon
   * @param x the x co-ordinate
   * @param y the y co-ordinate
   */
   private boolean gonContains(int x,int y)
   {   for(int z = 0;z < myGon.npoints;z++)
       {   if(myGon.xpoints[z] == x && myGon.ypoints[z] == y)
           {   Debug.O("polygon closed");
               return true;
           }
       }
       return false;
   }
   /** Fills up the polygon with the active plant, if the time has come for planting */
   private void fillGon()
   {   theLayers[this.curLayer].fillGon(this.activePlant,myGon);
   }
   /** if we are using the polygon, return the polygon.  Otherwise, return null
   * @return a polygon or null, depending on what the user is doing
   */
   public Polygon getPolygon()
   {   if(this.curAction == Constants.ACTION_PLANT_POLYGON)
       {   return this.myGon;
       }
       else
       {   return null;
       }
   }

   /** Update the information in one of the grid squares, and plant if necessary
    in the active layer of your planner.
    @param x the x coordinate
    @param y the y coordinate
    @param value the updated value in that square
    */
   public void updateData(int x, int y)
   {  Debug.O("clicked");
      if (this.isValidCoordinate(x, y))
      {	 this.lastSelection = this.action + " at (" + x + ", " + y + ").";

         // if plantingsquare is null, then this is the first time they've clicked, or they're not using plantingsquare, and are mulching or using the polygon
         if(this.plantingSquare == null || this.curAction == Constants.ACTION_QUERY || this.curAction == Constants.ACTION_PLANT_POLYGON)
         {   //updates planting square for the hell of it
             this.plantingSquare = new Rect(x,y);
             // if they're querying, we don't want plantingsquare around, and prepare the query
             if(this.curAction == Constants.ACTION_QUERY)
             {  this.prepareQuery(x,y);
                this.plantingSquare = null;
             }
             // if they're doing polygonal planting, then deal with it
             if(this.curAction == Constants.ACTION_PLANT_POLYGON)
             {  this.plantingSquare = null;
                if(this.myGon == null) // if they just started, or just drew one, then it will be null
                {   myGon = new Polygon();  // make a new polygon
                    myGon.addPoint(x,y); // add a point to it
                }
                else if(!gonContains(x,y)) // if the polygon already exists, and they haven't added a point that already exists
                {   myGon.addPoint(x,y);  // add another point
                }
                else
                {   this.saveGrids(); // if it isn't null, and does contain the new (x,y) point, then save the grids before we plant
                    fillGon(); // fills up this polygon on the layers
                    this.myGon = null; // clears out this polygon
                }

             } // end of if(plantingsquare == null)
         }
         else // if planting square isn't null, then they must be using it to do some sort of rectangular planting
         {  this.plantingSquare.setCorner(x+1,y+1); // sets the other corner of plantingsquare.
            if(x < this.plantingSquare.getLeft()) // these check to see if the corners are in different sectors then bottom right
            {   this.plantingSquare.setRight(this.plantingSquare.getLeft()+1);
                this.plantingSquare.setLeft(x);
            }
            if(y < this.plantingSquare.getTop())
            {   this.plantingSquare.setBottom(this.plantingSquare.getTop()+1);
                this.plantingSquare.setTop(y);
            }
            this.saveGrids(); // saves the grids before we change them, so we can undo
            if (this.curAction == Constants.ACTION_PLANT_NORMAL) // if they are planting normal rectangles...
            { this.placeStuff(this.activePlant,this.plantingSquare);
            }
            else if(this.curAction == Constants.ACTION_REMOVE) // if they are removing...
            { this.removeStuff(this.plantingSquare);
            }
            else if(this.curAction == Constants.ACTION_MULCH) // if they are mulching
            { this.addMulch(this.plantingSquare);
            }
            else if(this.curAction == Constants.ACTION_REM_MULCH) // if they are removing mulch
            { this.remMulch(this.plantingSquare);
            }
            else if(this.curAction == Constants.ACTION_TOPSOIL) // if they are adding topsoil
            { this.addTopsoil(this.plantingSquare);
            }
            else if(this.curAction == Constants.ACTION_REM_TOPSOIL) // if they are removing topsoil
            { this.remTopsoil(this.plantingSquare);
            }

            this.plantingSquare = null; // planting done, set this to null so it knows to restart at top-left again
            this.selectSquare = null; // planting done, set the highlighter square to null
         }
         this.myCanLayer = new CanopyLayer(theLayers);
         this.updateAllViews(); // hmm... I wonder what this does
      } // if
   } // updateData
   /** Adds topsoil to a given rectangle
   * @param where this is where we will add the topsoil to the plant layer
   */
   private void addTopsoil(Rect where)
   {  this.theLayers[Constants.PLANT_LAYER].setTopsoil(where,true);
   }
   /** Removes topsoil from a given rectangle
   * @param where this is where the topsoil will be removed from in the plant layer
   */
   private void remTopsoil(Rect where)
   {  this.theLayers[Constants.PLANT_LAYER].setTopsoil(where,false);
   }
   /** Adds mulch to a given rectangle in the plant layer
   * @param where where the mulch will be added in the plant layer
   */
   private void addMulch(Rect where)
   {  theLayers[Constants.PLANT_LAYER].setMulch(where, true);
   }
   /** Removes mulch from a given rectangle in the plant layer
   * @param where where the mulch will be removed from
   */
   private void remMulch(Rect where)
   {  theLayers[Constants.PLANT_LAYER].setMulch(where, false);
   }
   public void setDisplayHour(int whatHour)
   {   this.displayHour = whatHour;
       this.updateAllViews();
   }
   /** prepares the query string, for the feedback view to fetch later
   * @param x the x co-ord of the square to prepare it about
   * @param y the y co-ord of the square to prepare it about
   */
   private void prepareQuery(int x,int y)
   {   SquareContents sq = theLayers[this.curLayer].getSquare(x,y).getContents();
       Square mySquare = theLayers[this.curLayer].getSquare(x,y);
       char returnz = 13;
       this.curQuery = sq.getName() + "\n" + sq.getDesc() + "\nHas Water?:" + mySquare.getBesideWater() + "\nHas Mulch?" + mySquare.getMulch ()+ "\nHas Topsoil? " + mySquare.getTopsoil() + "\nEnvironmental Assessment: " + mySquare.getEnvironCost() + "\nShadow Size (squares): " + mySquare.getContents().getShadowSize() + "\nDeciduous?: " + mySquare.getContents().getIsDecid() + "/nShadow " + this.myCanLayer.getHoursOfSun (x,y);
   }
   /** returns the query prepared in the above method prepareQuery(int,int)
   * @return a string containing the query data
   */
   public String getQuery()
   {  return this.curQuery;
   }
   /** saves the model to a file
   * @param filename the filename where we will be saving it
   */
   public void save(String filename)
   {  TextOutput out = new TextOutput(filename);
      out.print(Constants.DEFAULT_WIDTH + " ");
      out.println(Constants.DEFAULT_HEIGHT);
      for(int x = 0;x < 2; x++)
      {  theLayers[x].save(out);
      }
      theLayers[Constants.PLANT_LAYER].saveMulch(out);

   }


	/** Return the last action performed.  This method is called from FeedbackView
	 * and will likely disappear as the project nears completion. */
   public String getLastSelection()
   {  return this.lastSelection;
   }


   // Methods involved in maintaining the views for the model

   /** Add a new view to your model.
    @param nextView a view which has just been constructed
    */
   public void addView(IView nextView)
   {  if (this.numViews == this.views.length)
      {  this.views = this.increaseSize();
      } // if
      this.views[numViews] = nextView;
      this.numViews++;
      nextView.updateView();
   } // addView

   /** @return a view list which is double the size of the current view list.
    */
   private IView[] increaseSize()
   {  IView[] temp = new IView[views.length * 2];
      for (int i = 0; i < this.views.length; i++)
      {  temp[i] = this.views[i];
      } // for
      return temp;
   } // increaseSize

   /** Refresh the views of the model after some new information has been
    acquired.
    */
   private void updateAllViews()
   {  for (int i = 0; i < this.numViews; i++)
      {  this.views[i].updateView();
      } // for
   } // updateAllViews

   /** sets the current action the landscape model will be doing
   * @param action the action that we will be doing
   */
   public void setAction(int action)
   {  this.curAction = action;
   }
   /** removes stuff from the layers.  If we are removing stuff from the base layer, remove stuff from the plant layer as well
   * @param where where we will be removing stuff from in curLayer
   */
   private void removeStuff(Rect where)
   {   if(this.curLayer == Constants.BASE_LAYER)
       {  this.theLayers[this.curLayer].setRect(where, squareList.getSquare(Constants.EMPTY_SQUARE), true);
          this.theLayers[Constants.PLANT_LAYER].setRect(where, squareList.getSquare(Constants.UNEDITABLE_SQUARE), true);
       }
       else if(this.curLayer == Constants.PLANT_LAYER)
       {  this.theLayers[this.curLayer].setRect(where, squareList.getSquare(Constants.EMPTY_SQUARE), true);
       }
       this.baseIsFull = theLayers[Constants.BASE_LAYER].isFull();
       this.calcWater();
       this.updateAllViews();
   }
   /** a helper method that goes through all the squares and calculates which ones are beside water
   */
   private void calcWater()
   {  Debug.O("calcing water");
      theLayers[Constants.BASE_LAYER].clearWater();
      theLayers[Constants.PLANT_LAYER].clearWater();

      for(int x = 0;x < Constants.DEFAULT_WIDTH;x++)
      {   for(int y = 0;y < Constants.DEFAULT_HEIGHT;y++)
          {   if(theLayers[Constants.BASE_LAYER].getSquare(x,y).getContents().equals(squareList.getSquare(Constants.NUM_WATER)))
              {   theLayers[Constants.BASE_LAYER].setWater(x,y,true);
                  theLayers[Constants.PLANT_LAYER].setWater(x,y,true);
              }
          }
      }
   }

} // LandscapePlanner
