package project.view;


import project.model.*;

import becker.util.IView;
import javax.swing.*;
import javax.swing.border.*;
import java.awt.*;
import java.awt.event.*;


/** A view of the planner which will represent the layout design.
 @author Art Hare, Waseem Quraishi and Kelly Leung */
public class GridView extends JPanel implements IView
{

   /** Constants for the graphics. */
   public static Color backgroundColour; //the background colour of the grid

   /** GUI widgets needed. */
   private JPanel grid = new JPanel();  //JPanel of the Grid
   private GridSection[][] layoutArea;  //The grid (2D array) of GridSetions

   /** The Landscape Design Planner application logic. */
   private LandscapePlanner model;  //a reference to the model
   private int mouseX;    //current x-coordinate position of mouse
   private int mouseY;    //current y-coordinate position of mouse
   private GridSection currentSection;  //current GridSection of the Grid
   private Rect selRect; // currently selected rectangle of squares
   private Rect lastSelRect;  //selected rectangle of squares from last cycle
   private Polygon lastGon; //selected polygon of squares from last cycle
   private Polygon curGon; // currently selected polygon of squares
   private boolean wasDrawingShadowLastTime;  // this keeps track of if we were drawing a shadow last update
   /** Construct the view of the planner which will represent the layout design.
    @param app the planner application logic
    */
   public GridView(LandscapePlanner app,int whichLayer)
   {  this.model = app;
      this.layoutView();
      this.registerListeners();
      this.model.addView(this);

   } // GridView

   /** Construct and layout the components of the view. */
   private void layoutView()
   {  int width = model.getWidth();
      int height = model.getHeight();

      this.layoutArea = new GridSection[width][height];
      this.setLayout(new GridLayout(height + 1, width + 1));

      this.fillGrid(width, height);
   } // layoutView

   // Helper methods for creating the layout.

   /** Construct and fill the 2-D grid which will display the design.
    @param width the width of the grid
    @param height the height of the grid
    */
   private void fillGrid(int width, int height)
   {  this.createXLabels(width);

      // Grid gridInfo = this.model.getGrid();
      for (int y = 0; y < height; y++)
      {  this.createYLabel(y);
         for (int x = 0; x < width; x++)
         {  Square data = this.model.getData(x, y);
            GridSection square = new GridSection(x, y, data,this);
            this.layoutArea[x][y] = square;
            square.setSize(GridSection.SQUARE_W, GridSection.SQUARE_H);
            this.add(square);
         } // for
      } // for
   } // fillGrid

   /** Construct the x-coordinate values which appear across the top of the grid
    @param width the width of the grid
    */
   private void createXLabels(int width)
   { // Create the empty space in the upper left corner
      JLabel xValue = new JLabel(" ", SwingConstants.CENTER);
      xValue.setSize(GridSection.SQUARE_W, GridSection.SQUARE_H);
      this.add(xValue);

      // Add the numbers across the top of the grid
      for (int x = 0; x < width; x++)
      {  xValue = new JLabel("" + x, SwingConstants.CENTER);
         xValue.setSize(GridSection.SQUARE_W, GridSection.SQUARE_H);
         this.add(xValue);
      } // for
   } // createXLabels

   /** Construct a y-coordinate value which appear on the left side of the grid
    @param yValue the y-coordinate value
    */
   private void createYLabel(int yValue)
   {  JLabel yLabel = new JLabel("" + yValue, SwingConstants.CENTER);
      yLabel.setSize(GridSection.SQUARE_W, GridSection.SQUARE_H);
      this.add(yLabel);
   } // createYLabel

   /** Create and attach listeners to appropriate components of the view. */
   private void registerListeners()
   {  GridSectionListener myListener = new GridSectionListener();
      for (int y = 0; y < this.model.getHeight(); y++)
      {  for (int x = 0; x < this.model.getWidth(); x++)
         {  this.layoutArea[x][y].addMouseListener(myListener);
            
         } // for
      } // for

   } // registerListeners

   /** Draws the objects selected from the actions and items drop menu, and draws the currently selected
       rectangle or polygon with the currently seleted item's background colour. */
   private void drawLayers()
   {   backgroundColour = this.getBackground();
    int width = this.model.getWidth();
   	int height = this.model.getHeight();
    for (int x = 0; x < width; x++)
    {  for (int y = 0; y < height; y++)
       {  GridSection square = layoutArea[x][y];
          Square data = this.model.getData(x, y);
          square.updateData(data,this.model);

       } // for
    } // for
    curGon = this.model.getPolygon();
    if(curGon == null || (this.lastGon != null && curGon == null)) // if the user isn't currently using the polygon, then do the standard rectangle drawing
    {
      selRect = this.model.getSelectSquare(); // gets the currently selected square
      if(lastSelRect != null && !lastSelRect.equals(selRect) && selRect != null) // if they have moved the rectangle, and both aren't null, then continue
      {
         Debug.O("painting away");

         for(int x = lastSelRect.getLeft();x <= lastSelRect.getRight();x++) // if there are squares not in the current selection
         {    for(int y = lastSelRect.getTop();y <= lastSelRect.getBottom();y++) // that were in the last selection, colour them grey
              {   if(!selRect.contains(x,y) && lastSelRect.contains(x,y))
                  {   this.layoutArea[x][y].setBackground(Constants.GRID_BACKGROUND);
                  }
             }
         }

         for(int x = selRect.getLeft();x <= selRect.getRight();x++) // if there are squares in the current section
         {   for(int y = selRect.getTop();y <= selRect.getBottom();y++) // that were not in the last section, colour them the
             {   if(!lastSelRect.contains(x,y) && selRect.contains(x,y))  // active colour.
                 {   this.layoutArea[x][y].setBackground(this.model.getActivePlant().getColor());
                 }
             }
         }

      }
      if(selRect == null && lastSelRect != null) // if the current selection is null, and the last one isn't, then they have just placed stuff.  We must clear out the last selection to grey
      {    for(int x = lastSelRect.getLeft();x <= lastSelRect.getRight();x++)
             {    for(int y = lastSelRect.getTop();y <= lastSelRect.getBottom();y++)
                  {   this.layoutArea[x][y].setBackground(Constants.GRID_BACKGROUND);
                  }
             }
      }
      if(selRect != null && lastSelRect == null) // if the current selection isn't null, and the last one is, they have just started placing stuff.  Draw everything
      {    for(int x = selRect.getLeft();x <= selRect.getRight();x++)
             {   for(int y = selRect.getTop();y <= selRect.getBottom();y++)
                 {   this.layoutArea[x][y].setBackground(this.model.getActivePlant().getColor());
                 }
             }
      }
      lastSelRect = selRect; // remembers what this selection was
    } // end of null check
    else // else if (polygon != null), essentially.  If the user is using the polygon, this is where things go
    {   this.curGon = this.model.getPolygon();
        if(this.lastGon != null && this.curGon == null) // if the current one is null, and the last one isn't, time for a clearing
        {   for(int x = 0;x < lastGon.npoints;x++)
            {   this.layoutArea[lastGon.xpoints[x]][lastGon.ypoints[x]].setBackground(Constants.GRID_BACKGROUND  );
                Debug.O("Resetting Colours at: " + lastGon.xpoints[x] + ", " + lastGon.ypoints[x]);
            }
        }
        if(this.curGon != null)
        {   for(int x = 0;x < curGon.npoints;x++)
            {   this.layoutArea[curGon.xpoints[x]][curGon.ypoints[x]].setBackground(this.model.getActivePlant().getColor());
            }
        }
        this.lastGon = this.curGon;
    }
   }
   /** Refresh the view with up-to-date information from the application logic. */
   public void updateView()
   {  Debug.O("redrawing gridview");

      if (this.model.getShowingShadow())
      {   this.drawShadow();
      }
      else
      {   if(this.wasDrawingShadowLastTime)
          {   this.clearGridBackgrounds();
          }
          this.drawLayers();
      }
      this.wasDrawingShadowLastTime = this.model.getShowingShadow();

   } // updateView

   /** Clears the background of grid to the regular colour of the grid */
   private void clearGridBackgrounds()
   {   for(int x = 0;x < this.layoutArea.length;x++)
       {   for(int y = 0;y < this.layoutArea[0].length;y++)
           {   this.layoutArea[x][y].setBackground(Constants.GRID_BACKGROUND );
           }
       }
   }

   /** Draws the shadow around the objects added to grid that cast shadows in
       surrounding areas */
   private void drawShadow()
   {  for(int x = 0;x < this.layoutArea.length;x++)
      {   for(int y =0;y < this.layoutArea[0].length;y++)
          {   int colorAtSquare = Constants.GREYSCALE * this.model.getSun(x,y);
              this.layoutArea[x][y].setBackground(new Color(colorAtSquare,colorAtSquare,colorAtSquare));
          }
      }
   }

   /** Sets the current GridSection the mouse is currently on
       @param x the given x position of the mouse
       @param y the given y position of the mouse */
   public void setCurrentSection(int x,int y)
   {
       if(this.currentSection != null && (x != this.currentSection.getCol() || y != this.currentSection.getRow()))
       {   this.currentSection.setBackground(Constants.GRID_BACKGROUND);
           this.model.setCurActiveSection(x,y);
           this.currentSection = this.layoutArea[x][y];

       }
       if(this.currentSection == null)
       {   this.currentSection = this.layoutArea[x][y];

       }
       //currentSection.setBackground(this.model.getActivePlant().getColor());

   }

   /** The interface will react to mouse actions. */
   private class GridSectionListener implements MouseListener
   { /** When the user clicks a grid section the application
       logic will be informed which grid section has just been chosen.
       */
      public void mouseClicked(MouseEvent evt)
      {
        GridSection lastSelected = (GridSection) evt.getSource();
        model.updateData(lastSelected.getCol(), lastSelected.getRow());
      } // mouseClicked

      public void mouseEntered(MouseEvent evt)
      {}

      public void mouseExited(MouseEvent evt)
      {}

      public void mousePressed(MouseEvent evt)
      {
      }
      public void mouseReleased(MouseEvent evt)
      {}
   } // GridSectionListener

} // GridView
